@@ -6,9 +6,17 @@
 functions = {}
 enums = {} # { enumType: { enumConstant: enumValue } }
 
+import os
 import sys
 import string
 
+if __name__ == "__main__":
+    # launched as a script
+    srcPref = os.path.dirname(sys.argv[0])
+else:
+    # imported
+    srcPref = os.path.dirname(__file__)
+
 #######################################################################
 #
 #  That part if purely the API acquisition phase from the
@@ -106,6 +114,7 @@ class docParser:
             self._data = []
             self.in_function = 1
             self.function = None
+            self.function_cond = None
             self.function_args = []
             self.function_descr = None
             self.function_return = None
@@ -114,6 +123,8 @@ class docParser:
                 self.function = attrs['name']
             if attrs.has_key('file'):
                 self.function_file = attrs['file']
+        elif tag == 'cond':
+            self._data = []
         elif tag == 'info':
             self._data = []
         elif tag == 'arg':
@@ -148,7 +159,7 @@ class docParser:
             if self.function != None:
                 function(self.function, self.function_descr,
                          self.function_return, self.function_args,
-                         self.function_file)
+                         self.function_file, self.function_cond)
                 self.in_function = 0
         elif tag == 'arg':
             if self.in_function == 1:
@@ -166,10 +177,16 @@ class docParser:
                 str = str + c
             if self.in_function == 1:
                 self.function_descr = str
+        elif tag == 'cond':
+            str = ''
+            for c in self._data:
+                str = str + c
+            if self.in_function == 1:
+                self.function_cond = str
                 
                 
-def function(name, desc, ret, args, file):
-    functions[name] = (desc, ret, args, file)
+def function(name, desc, ret, args, file, cond):
+    functions[name] = (desc, ret, args, file, cond)
 
 def enum(type, name, value):
     if not enums.has_key(type):
@@ -214,7 +231,7 @@ skipped_types = {
 py_types = {
     'void': (None, None, None, None),
     'int':  ('i', None, "int", "int"),
-    'long':  ('i', None, "int", "int"),
+    'long':  ('l', None, "long", "long"),
     'double':  ('d', None, "double", "double"),
     'unsigned int':  ('i', None, "int", "int"),
     'xmlChar':  ('c', None, "int", "int"),
@@ -270,6 +287,7 @@ py_types = {
     'xmlParserCtxt *': ('O', "parserCtxt", "xmlParserCtxtPtr", "xmlParserCtxtPtr"),
     'htmlParserCtxtPtr': ('O', "parserCtxt", "xmlParserCtxtPtr", "xmlParserCtxtPtr"),
     'htmlParserCtxt *': ('O', "parserCtxt", "xmlParserCtxtPtr", "xmlParserCtxtPtr"),
+    'xmlValidCtxtPtr': ('O', "ValidCtxt", "xmlValidCtxtPtr", "xmlValidCtxtPtr"),
     'xmlCatalogPtr': ('O', "catalog", "xmlCatalogPtr", "xmlCatalogPtr"),
     'FILE *': ('O', "File", "FILEPtr", "FILE *"),
     'xmlURIPtr': ('O', "URI", "xmlURIPtr", "xmlURIPtr"),
@@ -282,6 +300,9 @@ py_types = {
     'xmlRelaxNGPtr': ('O', "relaxNgSchema", "xmlRelaxNGPtr", "xmlRelaxNGPtr"),
     'xmlRelaxNGParserCtxtPtr': ('O', "relaxNgParserCtxt", "xmlRelaxNGParserCtxtPtr", "xmlRelaxNGParserCtxtPtr"),
     'xmlRelaxNGValidCtxtPtr': ('O', "relaxNgValidCtxt", "xmlRelaxNGValidCtxtPtr", "xmlRelaxNGValidCtxtPtr"),
+    'xmlSchemaPtr': ('O', "Schema", "xmlSchemaPtr", "xmlSchemaPtr"),
+    'xmlSchemaParserCtxtPtr': ('O', "SchemaParserCtxt", "xmlSchemaParserCtxtPtr", "xmlSchemaParserCtxtPtr"),
+    'xmlSchemaValidCtxtPtr': ('O', "SchemaValidCtxt", "xmlSchemaValidCtxtPtr", "xmlSchemaValidCtxtPtr"),
 }
 
 py_return_types = {
@@ -290,6 +311,20 @@ py_return_types = {
 
 unknown_types = {}
 
+foreign_encoding_args = (
+    'htmlCreateMemoryParserCtxt',
+    'htmlCtxtReadMemory',
+    'htmlParseChunk',
+    'htmlReadMemory',
+    'xmlCreateMemoryParserCtxt',
+    'xmlCtxtReadMemory',
+    'xmlCtxtResetPush',
+    'xmlParseChunk',
+    'xmlParseMemory',
+    'xmlReadMemory',
+    'xmlRecoverMemory',
+)
+
 #######################################################################
 #
 #  This part writes the C <-> Python stubs libxml2-py.[ch] and
@@ -297,11 +332,20 @@ unknown_types = {}
 #
 #######################################################################
 
+# Class methods which are written by hand in libxml.c but the Python-level
+# code is still automatically generated (so they are not in skip_function()).
+skip_impl = (
+    'xmlSaveFileTo',
+    'xmlSaveFormatFileTo',
+)
+
 def skip_function(name):
     if name[0:12] == "xmlXPathWrap":
         return 1
     if name == "xmlFreeParserCtxt":
         return 1
+    if name == "xmlCleanupParser":
+        return 1
     if name == "xmlFreeTextReader":
         return 1
 #    if name[0:11] == "xmlXPathNew":
@@ -309,6 +353,11 @@ def skip_function(name):
     # the next function is defined in libxml.c
     if name == "xmlRelaxNGFreeValidCtxt":
         return 1
+    if name == "xmlFreeValidCtxt":
+        return 1
+    if name == "xmlSchemaFreeValidCtxt":
+        return 1
+
 #
 # Those are skipped because the Const version is used of the bindings
 # instead.
@@ -333,6 +382,14 @@ def skip_function(name):
         return 1
     if name == "xmlErrMemory":
         return 1
+
+    if name == "xmlValidBuildContentModel":
+        return 1
+    if name == "xmlValidateElementDecl":
+        return 1
+    if name == "xmlValidateAttributeDecl":
+        return 1
+
     return 0
 
 def print_function_wrapper(name, output, export, include):
@@ -342,7 +399,7 @@ def print_function_wrapper(name, output, export, include):
     global skipped_modules
 
     try:
-        (desc, ret, args, file) = functions[name]
+        (desc, ret, args, file, cond) = functions[name]
     except:
         print "failed to get function %s infos"
         return
@@ -351,6 +408,9 @@ def print_function_wrapper(name, output, export, include):
         return 0
     if skip_function(name) == 1:
         return 0
+    if name in skip_impl:
+	# Don't delete the function entry in the caller.
+	return 1
 
     c_call = "";
     format=""
@@ -358,6 +418,7 @@ def print_function_wrapper(name, output, export, include):
     c_args=""
     c_return=""
     c_convert=""
+    num_bufs=0
     for arg in args:
         # This should be correct
         if arg[1][0:6] == "const ":
@@ -365,6 +426,8 @@ def print_function_wrapper(name, output, export, include):
         c_args = c_args + "    %s %s;\n" % (arg[1], arg[0])
         if py_types.has_key(arg[1]):
             (f, t, n, c) = py_types[arg[1]]
+	    if (f == 'z') and (name in foreign_encoding_args) and (num_bufs == 0):
+	        f = 't#'
             if f != None:
                 format = format + f
             if t != None:
@@ -375,6 +438,10 @@ def print_function_wrapper(name, output, export, include):
                    arg[1], t, arg[0]);
             else:
                 format_args = format_args + ", &%s" % (arg[0])
+	    if f == 't#':
+	        format_args = format_args + ", &py_buffsize%d" % num_bufs
+	        c_args = c_args + "    int py_buffsize%d;\n" % num_bufs
+		num_bufs = num_bufs + 1
             if c_call != "":
                 c_call = c_call + ", ";
             c_call = c_call + "%s" % (arg[0])
@@ -428,39 +495,10 @@ def print_function_wrapper(name, output, export, include):
             unknown_types[ret[0]] = [name]
         return -1
 
-    if file == "debugXML":
-        include.write("#ifdef LIBXML_DEBUG_ENABLED\n");
-        export.write("#ifdef LIBXML_DEBUG_ENABLED\n");
-        output.write("#ifdef LIBXML_DEBUG_ENABLED\n");
-    elif file == "HTMLtree" or file == "HTMLparser":
-        include.write("#ifdef LIBXML_HTML_ENABLED\n");
-        export.write("#ifdef LIBXML_HTML_ENABLED\n");
-        output.write("#ifdef LIBXML_HTML_ENABLED\n");
-    elif file == "c14n":
-        include.write("#ifdef LIBXML_C14N_ENABLED\n");
-        export.write("#ifdef LIBXML_C14N_ENABLED\n");
-        output.write("#ifdef LIBXML_C14N_ENABLED\n");
-    elif file == "xpathInternals" or file == "xpath":
-        include.write("#ifdef LIBXML_XPATH_ENABLED\n");
-        export.write("#ifdef LIBXML_XPATH_ENABLED\n");
-        output.write("#ifdef LIBXML_XPATH_ENABLED\n");
-    elif file == "xpointer":
-        include.write("#ifdef LIBXML_XPTR_ENABLED\n");
-        export.write("#ifdef LIBXML_XPTR_ENABLED\n");
-        output.write("#ifdef LIBXML_XPTR_ENABLED\n");
-    elif file == "xinclude":
-        include.write("#ifdef LIBXML_XINCLUDE_ENABLED\n");
-        export.write("#ifdef LIBXML_XINCLUDE_ENABLED\n");
-        output.write("#ifdef LIBXML_XINCLUDE_ENABLED\n");
-    elif file == "xmlregexp":
-        include.write("#ifdef LIBXML_REGEXP_ENABLED\n");
-        export.write("#ifdef LIBXML_REGEXP_ENABLED\n");
-        output.write("#ifdef LIBXML_REGEXP_ENABLED\n");
-    elif file == "xmlschemas" or file == "xmlschemastypes" or \
-         file == "relaxng":
-        include.write("#ifdef LIBXML_SCHEMAS_ENABLED\n");
-        export.write("#ifdef LIBXML_SCHEMAS_ENABLED\n");
-        output.write("#ifdef LIBXML_SCHEMAS_ENABLED\n");
+    if cond != None and cond != "":
+        include.write("#if %s\n" % cond)
+        export.write("#if %s\n" % cond)
+        output.write("#if %s\n" % cond)
 
     include.write("PyObject * ")
     include.write("libxml_%s(PyObject *self, PyObject *args);\n" % (name));
@@ -470,9 +508,17 @@ def print_function_wrapper(name, output, export, include):
 
     if file == "python":
         # Those have been manually generated
+	if cond != None and cond != "":
+	    include.write("#endif\n");
+	    export.write("#endif\n");
+	    output.write("#endif\n");
         return 1
     if file == "python_accessor" and ret[0] != "void" and ret[2] is None:
         # Those have been manually generated
+	if cond != None and cond != "":
+	    include.write("#endif\n");
+	    export.write("#endif\n");
+	    output.write("#endif\n");
         return 1
 
     output.write("PyObject *\n")
@@ -497,39 +543,10 @@ def print_function_wrapper(name, output, export, include):
     output.write(c_call)
     output.write(ret_convert)
     output.write("}\n\n")
-    if file == "debugXML":
-        include.write("#endif /* LIBXML_DEBUG_ENABLED */\n");
-        export.write("#endif /* LIBXML_DEBUG_ENABLED */\n");
-        output.write("#endif /* LIBXML_DEBUG_ENABLED */\n");
-    elif file == "HTMLtree" or file == "HTMLparser":
-        include.write("#endif /* LIBXML_HTML_ENABLED */\n");
-        export.write("#endif /* LIBXML_HTML_ENABLED */\n");
-        output.write("#endif /* LIBXML_HTML_ENABLED */\n");
-    elif file == "c14n":
-        include.write("#endif /* LIBXML_C14N_ENABLED */\n");
-        export.write("#endif /* LIBXML_C14N_ENABLED */\n");
-        output.write("#endif /* LIBXML_C14N_ENABLED */\n");
-    elif file == "xpathInternals" or file == "xpath":
-        include.write("#endif /* LIBXML_XPATH_ENABLED */\n");
-        export.write("#endif /* LIBXML_XPATH_ENABLED */\n");
-        output.write("#endif /* LIBXML_XPATH_ENABLED */\n");
-    elif file == "xpointer":
-        include.write("#endif /* LIBXML_XPTR_ENABLED */\n");
-        export.write("#endif /* LIBXML_XPTR_ENABLED */\n");
-        output.write("#endif /* LIBXML_XPTR_ENABLED */\n");
-    elif file == "xinclude":
-        include.write("#endif /* LIBXML_XINCLUDE_ENABLED */\n");
-        export.write("#endif /* LIBXML_XINCLUDE_ENABLED */\n");
-        output.write("#endif /* LIBXML_XINCLUDE_ENABLED */\n");
-    elif file == "xmlregexp":
-        include.write("#endif /* LIBXML_REGEXP_ENABLED */\n");
-        export.write("#endif /* LIBXML_REGEXP_ENABLED */\n");
-        output.write("#endif /* LIBXML_REGEXP_ENABLED */\n");
-    elif file == "xmlschemas" or file == "xmlschemastypes" or \
-         file == "relaxng":
-        include.write("#endif /* LIBXML_SCHEMAS_ENABLED */\n");
-        export.write("#endif /* LIBXML_SCHEMAS_ENABLED */\n");
-        output.write("#endif /* LIBXML_SCHEMAS_ENABLED */\n");
+    if cond != None and cond != "":
+        include.write("#endif /* %s */\n" % cond)
+        export.write("#endif /* %s */\n" % cond)
+        output.write("#endif /* %s */\n" % cond)
     return 1
 
 def buildStubs():
@@ -538,14 +555,14 @@ def buildStubs():
     global unknown_types
 
     try:
-	f = open("libxml2-api.xml")
+	f = open(os.path.join(srcPref,"libxml2-api.xml"))
 	data = f.read()
 	(parser, target)  = getparser()
 	parser.feed(data)
 	parser.close()
     except IOError, msg:
 	try:
-	    f = open("../doc/libxml2-api.xml")
+	    f = open(os.path.join(srcPref,"..","doc","libxml2-api.xml"))
 	    data = f.read()
 	    (parser, target)  = getparser()
 	    parser.feed(data)
@@ -559,7 +576,7 @@ def buildStubs():
 
     py_types['pythonObject'] = ('O', "pythonObject", "pythonObject", "pythonObject")
     try:
-	f = open("libxml2-python-api.xml")
+	f = open(os.path.join(srcPref,"libxml2-python-api.xml"))
 	data = f.read()
 	(parser, target)  = getparser()
 	parser.feed(data)
@@ -645,6 +662,7 @@ classes_type = {
     "xmlParserCtxt *": ("._o", "parserCtxt(_obj=%s)", "parserCtxt"),
     "htmlParserCtxtPtr": ("._o", "parserCtxt(_obj=%s)", "parserCtxt"),
     "htmlParserCtxt *": ("._o", "parserCtxt(_obj=%s)", "parserCtxt"),
+    "xmlValidCtxtPtr": ("._o", "ValidCtxt(_obj=%s)", "ValidCtxt"),
     "xmlCatalogPtr": ("._o", "catalog(_obj=%s)", "catalog"),
     "xmlURIPtr": ("._o", "URI(_obj=%s)", "URI"),
     "xmlErrorPtr": ("._o", "Error(_obj=%s)", "Error"),
@@ -656,6 +674,9 @@ classes_type = {
     'xmlRelaxNGPtr': ('._o', "relaxNgSchema(_obj=%s)", "relaxNgSchema"),
     'xmlRelaxNGParserCtxtPtr': ('._o', "relaxNgParserCtxt(_obj=%s)", "relaxNgParserCtxt"),
     'xmlRelaxNGValidCtxtPtr': ('._o', "relaxNgValidCtxt(_obj=%s)", "relaxNgValidCtxt"),
+    'xmlSchemaPtr': ("._o", "Schema(_obj=%s)", "Schema"),
+    'xmlSchemaParserCtxtPtr': ("._o", "SchemaParserCtxt(_obj=%s)", "SchemaParserCtxt"),
+    'xmlSchemaValidCtxtPtr': ("._o", "SchemaValidCtxt(_obj=%s)", "SchemaValidCtxt"),
 }
 
 converter_type = {
@@ -677,6 +698,9 @@ classes_ancestor = {
     "inputBuffer": "ioReadWrapper",
     "parserCtxt": "parserCtxtCore",
     "xmlTextReader": "xmlTextReaderCore",
+    "ValidCtxt": "ValidCtxtCore",
+    "SchemaValidCtxt": "SchemaValidCtxtCore",
+    "relaxNgValidCtxt": "relaxNgValidCtxtCore",
 }
 classes_destructors = {
     "parserCtxt": "xmlFreeParserCtxt",
@@ -689,17 +713,24 @@ classes_destructors = {
     "relaxNgSchema": "xmlRelaxNGFree",
     "relaxNgParserCtxt": "xmlRelaxNGFreeParserCtxt",
     "relaxNgValidCtxt": "xmlRelaxNGFreeValidCtxt",
+	"Schema": "xmlSchemaFree",
+	"SchemaParserCtxt": "xmlSchemaFreeParserCtxt",
+	"SchemaValidCtxt": "xmlSchemaFreeValidCtxt",
+        "ValidCtxt": "xmlFreeValidCtxt",
 }
 
 functions_noexcept = {
     "xmlHasProp": 1,
     "xmlHasNsProp": 1,
     "xmlDocSetRootElement": 1,
+    "xmlNodeGetNs": 1,
+    "xmlNodeGetNsDefs": 1,
 }
 
 reference_keepers = {
     "xmlTextReader": [('inputBuffer', 'input')],
     "relaxNgValidCtxt": [('relaxNgSchema', 'schema')],
+	"SchemaValidCtxt": [('Schema', 'schema')],
 }
 
 function_classes = {}
@@ -881,7 +912,7 @@ def buildWrappers():
 
     for name in functions.keys():
 	found = 0;
-	(desc, ret, args, file) = functions[name]
+	(desc, ret, args, file, cond) = functions[name]
 	for type in ctypes:
 	    classe = classes_type[type][2]
 
@@ -1006,6 +1037,13 @@ def buildWrappers():
 		classes.write("class %s(%s):\n" % (classname,
 			      classes_ancestor[classname]))
 		classes.write("    def __init__(self, _obj=None):\n")
+		if classes_ancestor[classname] == "xmlCore" or \
+		   classes_ancestor[classname] == "xmlNode":
+		    classes.write("        if type(_obj).__name__ != ")
+		    classes.write("'PyCObject':\n")
+		    classes.write("            raise TypeError, ")
+		    classes.write("'%s needs a PyCObject argument'\n" % \
+		                classname)
 		if reference_keepers.has_key(classname):
 		    rlist = reference_keepers[classname]
 		    for ref in rlist:
@@ -1017,7 +1055,7 @@ def buildWrappers():
 		   classes_ancestor[classname] == "xmlNode":
 		    classes.write("    def __repr__(self):\n")
 		    format = "<%s (%%s) object at 0x%%x>" % (classname)
-		    classes.write("        return \"%s\" %% (self.name, id (self))\n\n" % (
+		    classes.write("        return \"%s\" %% (self.name, long(pos_id (self)))\n\n" % (
 				  format))
 	    else:
 		txt.write("Class %s()\n" % (classname))
@@ -1029,17 +1067,25 @@ def buildWrappers():
 		        classes.write("        self.%s = None\n" % ref[1])
 		classes.write("        if _obj != None:self._o = _obj;return\n")
 		classes.write("        self._o = None\n\n");
+	    destruct=None
 	    if classes_destructors.has_key(classname):
 		classes.write("    def __del__(self):\n")
 		classes.write("        if self._o != None:\n")
 		classes.write("            libxml2mod.%s(self._o)\n" %
 			      classes_destructors[classname]);
 		classes.write("        self._o = None\n\n");
+		destruct=classes_destructors[classname]
 	    flist = function_classes[classname]
 	    flist.sort(functionCompare)
 	    oldfile = ""
 	    for info in flist:
 		(index, func, name, ret, args, file) = info
+		#
+		# Do not provide as method the destructors for the class
+		# to avoid double free
+		#
+		if name == destruct:
+		    continue;
 		if file != oldfile:
 		    if file == "python_accessor":
 			classes.write("    # accessors for %s\n" % (classname))
